%{
% Originally by Jakob Verbeek, Intelligent Systems Laboratory Amsterdam, 
% University of Amsterdam, 2005.
% http://www.science.uva.nl/~jverbeek

Modified by Charlie Tang
% 6/2011 CT: added some modifications so we can initialize differently
% params.PreWts.Psi - must be visible dim (D) by C. C is the number of mixtures
% params.PreWts.M - must be D by C - these are the means
% params.PreWts.W - D by nFactors by C - weights
% params.max_iters - limit number of iterations
% params.iso 
% params.eq
% params.epsilon - relative increase for convergence %default 1e-3
% params.GPU - use GPU or not
% if X has 3 dimensions, then we have a batched learning process

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================
%}

function [LogL,MoFA,Q] = mfa( d, C, params)
%
% [LogL,mfa,Q] = mfa(X,d,C,iso,dia);
% 
% xxTrain - global data matrix:
%           data matrix N x D x batches
% xxValid - N by D validation data (for early stopping)
% d     : latent dimensionality
% C     : number of mixtuer components
% iso   : use isotropic noise if 1
% eq    : use equal noise model for all components
% dia   : verbosity. 0=no output, 1=log-likelihood monitoring, 2=posterior plotting
%
% LogL  : row vector of log-likelihoods after each EM iteration
% MoFA  : structure containing the Mixture of Factor Analyzers
%         MoFA.W (D x d x C) : factor loading matrices
%         MoFA.M (D x C)     : component means
%         MoFA.Psi ( D x C ) : component noise variances
%         MoFA.mix ( C x 1 ) : mixing weights
% Q     : C x N matrix of component posterior probabilities
%
% by Jakob Verbeek, Intelligent Systems Laboratory Amsterdam, University of Amsterdam, 2005.
% http://www.science.uva.nl/~jverbeek

global xxTrain;
global xxValid;

if isfield( params, 'epsilon')
    epsilon = params.epsilon;
else
    epsilon     = 1e-3;  % relative log-likelihood increase threshold for convergence
end

if isfield( params, 'max_iters')
    max_iters = params.max_iters;
else
    max_iters = 1000; % maximum number of EM iterations
end
min_iters  = 5;       % minimum number of EM iterations

[N, D, Batches] = size(xxTrain);

if 1
    fprintf('\n--> MATLAB EM for MoFA\n');
	fprintf('    mixture components: %d\n',C);
	fprintf('    latent dimensions : %d\n',d);
	fprintf('    data: %d points in %d dimensions batches:%d\n', N, D, Batches);
	
    if params.iso==1
        fprintf(' !!! Using isotropic noise model\n');
    else
        fprintf('Non-Isotropic noise model\n');
    end
    if params.eq==1
        fprintf(' !!! Equal noise model in components\n');
    else
        fprintf('Non-equal noise model\n');
    end
end

% ----------- Initialization ---------------
if isfield( params.PreWts, 'M')
    MoFA.M = double(params.PreWts.M); 
else
    assert(false);
    %MoFA.M     = randperm(N); 
    %MoFA.M     = X(:,MoFA.M(1:C));    % means
end

if  isfield( params.PreWts, 'Psi')
    MoFA.Psi = double(params.PreWts.Psi);
else
    assert(false);
    %MoFA.Psi = 100*var(X(:))*ones(D,C);              % noise variance
end

if isfield( params.PreWts, 'W')
    MoFA.W = double(params.PreWts.W);
else
    assert(false);
    %MoFA.W = 0.01*randn(D,d,C);                      % factor loadings
end

if isfield( params.PreWts, 'mix')
    MoFA.mix   = params.PreWts.mix;                 % mixing weights
else     
    MoFA.mix   = ones (C,1)/C;  
end
% ----------- end ---------------

if exist('xxValid', 'var') && ~isempty(xxValid)
    likv = eval_mfa(double(xxValid)', MoFA);
    valid_logl = mean(likv);
    fprintf('\n*** Initial xxValid:%f ***\n', valid_logl);
else
    valid_logl = -inf;
end

LogL=zeros(1,max_iters);

numbatchesUse = min(Batches, params.num_mbs);
inds = randperm( Batches );
inds = inds(1:numbatchesUse);
    
for bb = 1:length(inds) %loop thru minibatches    
    for iter=1:max_iters; % EM loops
        
        mstep_p =[];
        mstep_p.eq = params.eq;
        mstep_p.iso = params.iso;
        mstep_p.GPU = params.GPU;
        mstep_p.QTHRESH = params.QTHRESH;
        mstep_p.MIN_PSI = params.MIN_PSI;
        
        X = xxTrain(:,:,inds(bb));
        X2 = X.^2;
     
        [Q, LL, Lats, lat_covs] = mfa_E_step( double(X)', MoFA, double(X2)' ); % E-step
        LogL(iter) = sum(LL);        
        MoFA_new = mfa_M_step(mstep_p, double(X)', double(X2)', Q, Lats, lat_covs, MoFA, params.ZM);  % M-step
        
        clear Lats;        
        if exist('xxValid', 'var') && ~isempty(xxValid)
            likv = eval_mfa(double(xxValid)', MoFA_new);
            valid_logl_new = mean(likv);            
            fprintf(' xxValid:%f ', valid_logl_new);
            
            if valid_logl_new < valid_logl
                fprintf('--> Early Stopping! %d, \n', iter );
                break;
            end
            valid_logl = valid_logl_new;
            
        end
        
        if iter > 1
            rel_change = (LogL(iter)-LogL(iter-1)) / abs(mean(LogL(iter-1:iter)));
            
            if rel_change < 0
                fprintf('--> Log likelihood decreased in iteration %d,  relative change %f\n',iter,rel_change);
                break;
            end
            
            MoFA = MoFA_new;

            fprintf('iteration %3d minibatches:%d/%d avgLL:%.2f Logl %.2f  relative increment  %.6f\n',iter, bb, numbatchesUse, LogL(iter)/N, LogL(iter),rel_change);

            %if dia>1; clf;set(gcf,'Double','on');imagesc(Q);xlabel('data point index');ylabel('mixture component index');set(gca,'ytick',1:C);colormap gray;colorbar;title('Posterior distributions on mixture components given data point');drawnow;end
            if (rel_change < epsilon) && (iter > min_iters); 
                break;
            end

        elseif iter == 1
            MoFA = MoFA_new;
            fprintf('iteration %3d  avgLL:%.2f Logl %.2f \n',iter, LogL(iter)/N, LogL(iter));
        end         
    end
end
